///***************************************************************************
// File:		USZMenneBiomed.c
///***************************************************************************

//
// Includes
//
#include <windows.h>
#include "xlcall.h"
#include "framewrk.h"
#include <math.h>
#include <float.h>
//
// Later, the instance handle is required to create dialog boxes.
// g_hInst holds the instance handle passed in by DllMain so that it is
// available for later use. 
// Global Variables
//
HANDLE g_hInst = NULL;
//
// Syntax of the Register Command:
// REGISTER(module_text, procedure, type_text, function_text, 
//          argument_text, macro_type, category, shortcut_text,
//          help_topic, function_help, argument_help1, argument_help2,...)
//
//
// g_rgWorksheetFuncs will use only the first 11 arguments of 
//                  the Register function.
//
// This is a table of all the worksheet functions exported by this module.
// These functions are all registered (in xlAutoOpen) when you
// open the XLL. Before every string, leave a space for the
// byte count. The format of this table is the same as 
// arguments two through eleven of the REGISTER function.
// g_rgWorksheetFuncsRows define the number of rows in the table. The
// g_rgWorksheetFuncsCols represents the number of columns in the table.
//
#define g_rgWorksheetFuncsRows 1
#define g_rgWorksheetFuncsCols 10

static LPSTR g_rgWorksheetFuncs
[g_rgWorksheetFuncsRows][g_rgWorksheetFuncsCols] =
{
  { " LinExpT50",                          // Procedure
    " BBB",                                // Out and 2 in parameters double
    " LinExpT50",                          // function_text
    " kappa,t(empt)",                      // argument_text
    " 1",                                  // macro_type
    " USZ MenneBiomed",                    // category
    " ",                                   // shortcut_text
    " ",                                   // help_topic
    " Returns T(50%) for LinExp model given kappa and t(empt)",   // function_help
    " (1+kappa*t50/tempt)*exp(-t50/tempt)=0.5"    // argument_help1
  },
};


///***************************************************************************
// DllMain()
//
// Purpose:
//
//   Windows 95 and Windows NT call one function, DllMain, for both 
//   initialization and termination. It also makes calls on both a 
//   per-process and per-thread basis, so several initialization calls
//   can be made if a process is multithreaded.
//   This function is called when the DLL is first loaded, with a dwReason
//   of DLL_PROCESS_ATTACH. In this example, we byte-count all the strings
//   in the preceding tables.
//
// Parameters:
//
//        HANDLE hDLL       - Module handle.
//        DWORD dwReason,   - Reason for call
//        LPVOID lpReserved - Reserved
//
// Returns: The function returns TRUE (1) to indicate success. If, during
//          per-process initialization, the function returns zero, 
//          the system cancels the process.
//
// Comments:
//
// History:  Date       Author        Reason
///***************************************************************************

BOOL APIENTRY DllMain( HANDLE hDLL, 
                      DWORD dwReason, 
                      LPVOID lpReserved )
{
  switch (dwReason)
  {
  case DLL_PROCESS_ATTACH:
    {
      //
      // In the following five loops, the strings in g_rgWorksheetFuncs[], 
      // g_rgCommandFuncs[], g_rgTool[], g_rgMenu[] and g_rgDialog[] are 
      // byte-counted so that they won't need to be byte-counted later.
      // In addition, the instance handle passed into DllMain is saved
      // in the global variable g_hInst for later use.
      //
      int i,j;

      g_hInst = hDLL;

      for (i=0; i<g_rgWorksheetFuncsRows; i++) 
      {
        for (j=0; j<g_rgWorksheetFuncsCols; j++) 
        {
          g_rgWorksheetFuncs[i][j][0] = 
            (BYTE) lstrlen (g_rgWorksheetFuncs[i][j]+1);
        }
      }

      break;
    }
  case DLL_PROCESS_DETACH:
  case DLL_THREAD_ATTACH:
  case DLL_THREAD_DETACH:
  default:
    break;
  }
  return TRUE;
}


///***************************************************************************
// xlAutoOpen()
//
// Purpose: Microsoft Excel call this function when the DLL is loaded.
//
//  Microsoft Excel uses xlAutoOpen to load XLL files.
//  When you open an XLL file, the only action
//  Microsoft Excel takes is to call the xlAutoOpen function.
//
//  More specifically, xlAutoOpen is called:
//
//   - when you open this XLL file from the File menu,
//   - when this XLL is in the XLSTART directory, and is
//     automatically opened when Microsoft Excel starts,
//   - when Microsoft Excel opens this XLL for any other reason, or
//   - when a macro calls REGISTER(), with only one argument, which is the
//     name of this XLL.
//
//  xlAutoOpen is also called by the Add-in Manager when you add this XLL 
//  as an add-in. The Add-in Manager first calls xlAutoAdd, then calls
//  REGISTER("EXAMPLE.XLL"), which in turn calls xlAutoOpen.
//
//  xlAutoOpen should:
//
//   - register all the functions you want to make available while this
//     XLL is open,
//
//   - add any menus or menu items that this XLL supports,
//
//   - perform any other initialization you need, and
//
//   - return 1 if successful, or return 0 if your XLL cannot be opened.
//
// Parameters: None
//
// Returns: 1 if successful, or 0 if the XLL cannot be opened.
//
// Comments:
//
// History:  Date       Author        Reason
///***************************************************************************

__declspec(dllexport) int WINAPI xlAutoOpen(void)
{

  static XLOPER xDLL,    // name of this DLL //
    xMenu,   // xltypeMulti containing the menu //
    xTool,   // xltypeMulti containing the toolbar //
    xTest;   // used for menu test //
  int i;              // Loop indices //

  //
  // In the following block of code the name of the XLL is obtained by
  // calling xlGetName. This name is used as the first argument to the
  // REGISTER function to specify the name of the XLL. Next, the XLL loops
  // through the g_rgWorksheetFuncs[] table, and the g_rgCommandFuncs[]
  // tableregistering each function in the table using xlfRegister. 
  // Functions must be registered before you can add a menu item.
  //
  HMODULE hmodDLL1,hmodDLL2;
  char szFile1[256], szFile2[256];
  DWORD dwRet; 

  hmodDLL1 = GetModuleHandle("USZMenneBiomed.dll");
  hmodDLL2 = GetModuleHandle(NULL);

  dwRet = GetModuleFileName(hmodDLL1, szFile1, 256);
  dwRet = GetModuleFileName(hmodDLL2, szFile2, 256);

  Excel(xlGetName, &xDLL, 0);

  for (i=0; i<g_rgWorksheetFuncsRows; i++) 
  {
    Excel(xlfRegister, 0, 1+ g_rgWorksheetFuncsCols,
      (LPXLOPER) &xDLL,
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][0]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][1]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][2]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][3]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][4]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][5]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][6]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][7]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][8]),
      (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][9]));
  }

  // Free the XLL filename //
  Excel(xlFree, 0, 2, (LPXLOPER) &xTest, (LPXLOPER) &xDLL);
  return 1;
}


///***************************************************************************
// xlAutoClose()
//
// Purpose: Microsoft Excel call this function when the DLL is unloaded.
//
//  xlAutoClose is called by Microsoft Excel:
//
//   - when you quit Microsoft Excel, or
//   - when a macro sheet calls UNREGISTER(), giving a string argument
//     which is the name of this XLL.
//
//  xlAutoClose is called by the Add-in Manager when you remove this XLL from
//  the list of loaded add-ins. The Add-in Manager first calls xlAutoRemove,
//  then calls UNREGISTER("EXAMPLE.XLL"), which in turn calls xlAutoClose.
// 
//  xlAutoClose is called by EXAMPLE.XLL by the function fExit. This function
//  is called when you exit Example.
// 
//  xlAutoClose should:
// 
//   - Remove any menus or menu items that were added in xlAutoOpen,
// 
//   - do any necessary global cleanup, and
// 
//   - delete any names that were added (names of exported functions, and 
//     so on). Remember that registering functions may cause names to 
//     be created.
// 
//  xlAutoClose does NOT have to unregister the functions that were registered
//  in xlAutoOpen. This is done automatically by Microsoft Excel after
//  xlAutoClose returns.
// 
//  xlAutoClose should return 1.
//
// Parameters: None
//
// Returns: 1
//
// Comments:
//
// History:  Date       Author        Reason
///***************************************************************************

__declspec(dllexport) int WINAPI xlAutoClose(void)
{
  int i;
  for (i = 0; i < g_rgWorksheetFuncsRows; i++)
    Excel(xlfSetName, 0, 1, TempStr(g_rgWorksheetFuncs[i][2]));  
  return 1;
}


///***************************************************************************
// lpstricmp()
//
// Purpose: Compares two Pascal strings and checks if they are equal,
//          without case sensitivity. This routine is useful for
//          xlAutoRegister         
//
// Parameters:
//
//      LPSTR s     First string
//      LPSTR t     Second string
//
// Returns: 
//
//      int         0 if they are equal
//                  Nonzero otherwise
//
// Comments:
//
//      Unlike the usual string functions, lpstricmp
//      doesn't care about collating sequence.
//
// History:  Date       Author        Reason
///***************************************************************************

int lpstricmp(LPSTR s, LPSTR t)
{
  int i;

  if (*s != *t)
    return 1;

  for (i = 1; i <= s[0]; i++) 
  {
    if (tolower(s[i]) != tolower(t[i]))
      return 1;
  }
  return 0;
}


///***************************************************************************
// xlAutoRegister()
//
// Purpose:
//
//   This function is called by Microsoft Excel if a macro sheet tries to
//   register a function without specifying the type_text argument. If that
//   happens, Microsoft Excel calls xlAutoRegister, passing the name of the
//   function that the user tried to register. xlAutoRegister should use the
//   normal REGISTER function to register the function, only this time it must
//   specify the type_text argument. If xlAutoRegister does not recognize the
//   function name, it should return a #VALUE! error. Otherwise, it should
//   return whatever REGISTER returned.
//
// Parameters:
//
//      LPXLOPER pxName     xltypeStr containing the
//                          name of the function
//                          to be registered. This is not
//                          case sensitive.
//
// Returns: 
//
//      LPXLOPER            xltypeNum containing the result
//                          of registering the function,
//                          or xltypeErr containing #VALUE!
//                          if the function could not be
//                          registered.
//
// Comments:
//
// History:  Date       Author        Reason
///***************************************************************************

__declspec(dllexport) LPXLOPER WINAPI xlAutoRegister(LPXLOPER pxName)
{
  static XLOPER xDLL, xRegId;
  int i;

  //
  // This block initializes xRegId to a #VALUE! error first. This is done in
  // case a function is not found to register. Next, the code loops through 
  // the functions in g_rgFuncs[] and uses lpstricmp to determine if the 
  // current row in g_rgFuncs[] represents the function that needs to be 
  // registered. When it finds the proper row, the function is registered 
  // and the register ID is returned to Microsoft Excel. If no matching 
  // function is found, an xRegId is returned containing a #VALUE! error.
  //

  xRegId.xltype = xltypeErr;
  xRegId.val.err = xlerrValue;


  for (i=0; i<g_rgWorksheetFuncsRows; i++) 
  {
    if (!lpstricmp(g_rgWorksheetFuncs[i][0], pxName->val.str))	
    {
      Excel(xlfRegister, 0, 1+ g_rgWorksheetFuncsCols,
        (LPXLOPER) &xDLL,
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][0]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][1]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][2]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][3]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][4]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][5]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][6]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][7]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][8]),
        (LPXLOPER) TempStr(g_rgWorksheetFuncs[i][9]));
      /// Free oper returned by xl //
      Excel(xlFree, 0, 1, (LPXLOPER) &xDLL);

      return (LPXLOPER) &xRegId;
    }
  }

  return (LPXLOPER) &xRegId;
}

///***************************************************************************
// xlAutoAdd()
//
// Purpose:
//
//    This function is called by the Add-in Manager only. When you add a
//    DLL to the list of active add-ins, the Add-in Manager calls xlAutoAdd()
//    and then opens the XLL, which in turn calls xlAutoOpen.
//
// Parameters:
//
// Returns: 
//
//      int       1
//
// Comments:
//
// History:  Date       Author        Reason
///***************************************************************************

__declspec(dllexport) int WINAPI xlAutoAdd(void)
{                                 
  char szBuf[255];

  wsprintf((LPSTR)szBuf, " Thank you for adding USZMenneBiomed.XLL\n "
    "built on %s at %s", __DATE__, __TIME__);

  // Display a dialog box indicating that the XLL was successfully added //
  Excel(xlcAlert, 0, 2, TempStr(szBuf), TempInt(2));
  return 1;
}

///***************************************************************************
// xlAutoRemove()
//
// Purpose:
//
//    This function is called by the Add-in Manager only. When you remove
//    an XLL from the list of active add-ins, the Add-in Manager calls
//    xlAutoRemove() and then UNREGISTER("EXAMPLE.XLL").
//   
//    You can use this function to perform any special tasks that need to be
//    performed when you remove the XLL from the Add-in Manager's list
//    of active add-ins. For example, you may want to delete an
//    initialization file when the XLL is removed from the list.
///***************************************************************************

__declspec(dllexport) int WINAPI xlAutoRemove(void)
{
  // Show a dialog box indicating that the XLL was successfully removed //
  Excel(xlcAlert, 0, 2, TempStr(" Farewell to USZMenneBiomed.XLL!"),
    TempInt(2));
  return 1;
}

///***************************************************************************
// xlAddInManagerInfo()
//
// Purpose:
//
//    This function is called by the Add-in Manager to find the long name
//    of the add-in. If xAction = 1, this function should return a string
//    containing the long name of this XLL, which the Add-in Manager will use
//    to describe this XLL. If xAction = 2 or 3, this function should return
//    #VALUE!.
//
// Parameters:
//
//      LPXLOPER xAction    What information you want. One of:
//                          1 = the long name of the
//                              add-in
//                          2 = reserved
//                          3 = reserved
// Returns: 
//
//      LPXLOPER            The long name or #VALUE!.
//

__declspec(dllexport) LPXLOPER WINAPI xlAddInManagerInfo(LPXLOPER xAction)
{
  static XLOPER xInfo, xIntAction;
  Excel(xlCoerce, &xIntAction, 2, xAction, TempInt(xltypeInt));
  if(xIntAction.val.w == 1) 
  {
    xInfo.xltype = xltypeStr;
    xInfo.val.str = "\015MenneBiomed USZ" ;
  }
  else 
  {
    xInfo.xltype = xltypeErr;
    xInfo.val.err = xlerrValue;
  }

  return (LPXLOPER) &xInfo;
}




typedef struct {
  double kappa,tempt;
}LinExpPar;

double LinExp(double t, void* p)
{
  LinExpPar * par=p;
  return (1+par->kappa*t/par->tempt)*(exp(-t/par->tempt))-0.5;
}


double zeroin(			/* An estimate of the root */
              double ax,				/* Left border | of the range	*/
              double bx,				/* Right border| the root is seeked*/
              double (*f)(double x, void *info),	/* Function under investigation	*/
              void *info,				/* Add'l info passed on to f	*/
              double *Tol,			/* Acceptable tolerance		*/
              int *Maxit)				/* Max # of iterations */
{
  double a,b,c,			/* Abscissae, descr. see above	*/
    fa, fb, fc;			/* f(a), f(b), f(c) */
  double tol;
  int maxit;

  a = ax;  b = bx;  fa = (*f)(a, info);  fb = (*f)(b, info);
  c = a;   fc = fa;
  maxit = *Maxit + 1; tol = * Tol;

  while(maxit--)		/* Main iteration loop	*/
  {
    double prev_step = b-a;		/* Distance from the last but one
                              to the last approximation	*/
    double tol_act;			/* Actual tolerance		*/
    double p;			/* Interpolation step is calcu- */
    double q;			/* lated in the form p/q; divi-
                  * sion operations is delayed
                  * until the last moment	*/
    double new_step;		/* Step at this iteration	*/

    if( fabs(fc) < fabs(fb) )
    {				/* Swap data for b to be the	*/
      a = b;  b = c;  c = a;	/* best approximation		*/
      fa=fb;  fb=fc;  fc=fa;
    }
    tol_act = 2*DBL_EPSILON*fabs(b) + tol/2;
    new_step = (c-b)/2;

    if( fabs(new_step) <= tol_act || fb == (double)0 )
    {
      *Maxit -= maxit;
      *Tol = fabs(c-b);
      return b;			/* Acceptable approx. is found	*/
    }

    /* Decide if the interpolation can be tried	*/
    if( fabs(prev_step) >= tol_act	/* If prev_step was large enough*/
      && fabs(fa) > fabs(fb) ) {	/* and was in true direction,
                                  * Interpolation may be tried	*/
        register double t1,cb,t2;
        cb = c-b;
        if( a==c ) {		/* If we have only two distinct	*/
          /* points linear interpolation	*/
          t1 = fb/fa;		/* can only be applied		*/
          p = cb*t1;
          q = 1.0 - t1;
        }
        else {			/* Quadric inverse interpolation*/

          q = fa/fc;  t1 = fb/fc;	 t2 = fb/fa;
          p = t2 * ( cb*q*(q-t1) - (b-a)*(t1-1.0) );
          q = (q-1.0) * (t1-1.0) * (t2-1.0);
        }
        if( p>(double)0 )		/* p was calculated with the op-*/
          q = -q;			/* posite sign; make p positive	*/
        else			/* and assign possible minus to	*/
          p = -p;			/* q				*/

        if( p < (0.75*cb*q-fabs(tol_act*q)/2) /* If b+p/q falls in [b,c]*/
          && p < fabs(prev_step*q/2) )	/* and isn't too large	*/
          new_step = p/q;			/* it is accepted
                              * If p/q is too large then the
                              * bisection procedure can
                              * reduce [b,c] range to more
                              * extent */
      }

      if( fabs(new_step) < tol_act) {	/* Adjust the step to be not less*/
        if( new_step > (double)0 )	/* than tolerance		*/
          new_step = tol_act;
        else
          new_step = -tol_act;
      }
      a = b;	fa = fb;			/* Save the previous approx. */
      b += new_step;	fb = (*f)(b, info);	/* Do step to a new approxim. */
      if( (fb > 0 && fc > 0) || (fb < 0 && fc < 0) ) {
        /* Adjust c for it to have a sign opposite to that of b */
        c = a;  fc = fa;
      }

  }
  /* failed! */
  *Tol = fabs(c-b);
  return b;
}



__declspec(dllexport) double WINAPI LinExpT50( double kappa, double tempt) 
{
  int MaxIt=100;
  double Tol=1.e-4;
  LinExpPar par={kappa,tempt};
  return zeroin(0,10*tempt,LinExp,(void *)&par,&Tol,&MaxIt);
}